# implement output in OFX 1 format.

package OFX::OFX1;

require 5.10.0;
use Carp;

# generate OFX 1 header string.
#
# the parameter hash can contain the following optional keys:
# encoding - "USASCII" or "UNICODE".
# charset - character set string.
#
# @param $h reference to hash.
# @return OFX 1.0.3 header string.
sub header
{
  my ($h) = @_;
  $h = {} unless ref($h) eq 'HASH';
  my $encoding = $h->{encoding};
  $encoding = 'USASCII' unless $encoding;

  my $charset = $h->{charset};
  $charset = '1252' unless $charset;

  my $a = <<"EOT";
OFXHEADER:100
DATA:OFXSGML
VERSION:103
SECURITY:NONE
ENCODING:$encoding
CHARSET:$charset
COMPRESSION:NONE
OLDFILEUID:NONE
NEWFILEUID:NONE

EOT
  return $a;
}

# create a string with a <STATUS> tag.
#
# the hash reference parameter can contain the following optional key/value pairs:
# code - integer code value.
# severity - severity string.
# message - status description string.
#
# @return string.
sub STATUS
{
  my ($h) = @_;
  $h = {} unless ref($h) eq 'HASH';

  my $code = $h->{code};
  $code = 0 unless $code;

  my $severity = $h->{severity};
  $severity = (($code > 1) ? 'ERROR' : 'INFO') unless $severity;

  my $message  = "<MESSAGE>$h->{message}" if length($h->{message});

  my $a = <<"EOT";
<STATUS>
  <CODE>$code
  <SEVERITY>$severity
  $message
</STATUS>
EOT
  return $a;
};

# create a string with a <FI> tag.
#
# the hash reference parameter has to contain the following key/value pairs:
# org - organization string.
#
# the hash reference parameter can contain the following optional key/value pairs:
# fid - financial institution ID string.
#
# if the hash reference parameter is undef, an empty string is returned.
#
# @return string.
sub FI
{
  my ($h) = @_;
  return '' unless ref($h) eq 'HASH';
  croak("FI::ORG not set") unless $h->{org};
  my $fid = "<FID>$h->{fid}" if length($h->{fid});

  my $a = <<"EOT";
<FI>
  <ORG>$h->{org}
  $fid
</FI>
EOT
  return $a;
}

# create a datetime string in GMT.
# @param number of seconds since epoch. If undef use current system time.
# @return datetime string.
sub datetime
{
  my ($ts) = @_;
  $ts = time() unless $ts;
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($ts);
  return sprintf('%04i%02i%02i%02i%02i%02i.000[0:GMT]', $year+1900, $mon+1, $mday, $hour, $min, $sec);
}

# create a string with a <SIGNONMSGSRSV1> tag.
#
# the hash reference parameter can contain the following optional key/value pairs:
# status - reference to a STATUS hash.
# dtserver - time stamp string when server created response, in seconds since epoch.
# language - language string.
# dtprofup - time stamp string when profile was updated.
# dtacctup - time stamp string when account was updated.
# fi - reference to a FI hash.
#
# @return string.
sub SIGNONMSGSRSV1
{
  my ($h) = @_;
  return '' unless ref($h) eq 'HASH';

  my $dtserver = datetime($h->{dtserver});

  my $language = $h->{language};
  $language = 'ENG' unless $language;

  my $dtprofup = "<DTPROFUP>$h->{dtprofup}" if $h->{dtprofup};
  my $dtacctup = "<DTACCTUP>$h->{dtacctup}" if $h->{dtacctup};

  my $status = STATUS($h->{status});
  my $fi = FI($h->{fi});

  my $a = <<"EOT";
<SIGNONMSGSRSV1>
  <SONRS>
$status
  <DTSERVER>$dtserver
  <LANGUAGE>$language
  $dtprofup
  $dtacctup
$fi
  </SONRS>
</SIGNONMSGSRSV1>
EOT
  return $a;
}

# create a string with a <BANKACCTFROM> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# bankid - routing and transit number string.
# acctid - account number string.
# accttype - account type string, valid values: "CHECKING", "SAVINGS", "MONEYMRKT", "CREDITLINE".
#
# the hash reference parameter can contain the following optional key/value pairs:
# branchid - bank identifier for international banks string.
# acctkey - checksum for international banks string.
#
# @return string
sub BANKACCTFROM
{
  my ($h) = @_;
  croak("BANKACCTFROM::bankid not found") unless $h->{bankid};
  croak("BANKACCTFROM::acctid not found") unless $h->{acctid};
  croak("BANKACCTFROM::accttype not found") unless $h->{accttype};
  croak("BANKACCTFROM::accttype not valid") unless $h->{accttype} ~~ [qw{CHECKING SAVINGS MONEYMRKT CREDITLINE}];

  my $branchid = "<BRANCHID>$h->{branchid}" if length($h->{branchid});
  my $acctkey = "<ACCTKEY>$h->{acctkey}" if length($h->{acctkey});

  my $a = <<"EOT";
<BANKACCTFROM>
  <BANKID>$h->{bankid}
  $branchid
  <ACCTID>$h->{acctid}
  <ACCTTYPE>$h->{accttype}
  $acctkey
</BANKACCTFROM>
EOT
  return $a;
}

# create a string with a <STMTTRN> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# trntype - transaction type string, valid values: qw(CREDIT DEBIT INT DIV FEE SRVCHG DEP ATM POS XFER CHECK PAYMENT CASH DIRECTDEP DIRECTDEBIT REPEATPMT OTHER).
# dtposted - timestamp when transaction was posted, in seconds since epoch.
# trnamt - amount of transaction.
# fitid - transaction ID string issued by financial institution.
#
# the hash reference parameter can contain the following optional key/value pairs:
# dtuser - timestamp when transaction was initiated, in seconds since epoch.
# dtavail - timestamp when funds are available, in seconds since epoch.
# correctfitid - transaction ID of a previous transaction that is corrected.
# correctaction - valid values: "REPLACE", "DELETE".
# srvrtid - server assigned transaction ID string.
# checknum - check number.
# refnum - reference number.
# sic - standard industrial code.
# payeeid - payee identifier.
# name - name of payee or description of transaction.
# memo - extra information, not in 'name'.
#
# \todo support tags <BANKACCTTO>, <CCACCTTO>, <CURRENCY>, <ORIGCURRENCY>, <PAYEE>.
#
# @return string.
sub STMTTRN
{
  my ($h) = @_;
  croak("STMTTRN is not a hash reference") unless ref($h) eq 'HASH';
  croak("STMTTRN::trntype is not valid") unless $h->{trntype} ~~ [qw(CREDIT DEBIT INT DIV FEE SRVCHG DEP ATM POS XFER CHECK PAYMENT CASH DIRECTDEP DIRECTDEBIT REPEATPMT OTHER)];
  croak("STMTTRN::dtposted is not valid") unless $h->{dtposted} > 0;
  croak("STMTTRN::trnamt not found") unless length($h->{trnamt});
  croak("STMTTRN::fitid not found") unless length($h->{fitid});
  if ($h->{correctaction}) {
    croak("STMTTRN::correctaction is invalid") unless $h->{correctaction} ~~ [qw(REPLACE DELETE)];
  }

  # check that transaction type matches sign of transaction amount
  if ($h->{trntype} ~~ [qw(CREDIT DIV DEP DIRECTDEP)]) {
     croak("STMTTRN::trnamt must be positive for credit type transaction") unless $h->{trnamt} >= 0.0;
  } elsif ($h->{trntype} ~~ [qw(DEBIT FEE SRVCHG CHECK CASH DIRECTDEBIT)]) {
    croak("STMTTRN::trnamt must be negative for debit type transaction") unless $h->{trnamt} <= 0.0;
  } elsif ($h->{trntype} ~~ [qw(INT ATM POS OTHER XFER PAYMENT REPEATPMT)]) {
    # can have any sign.
  } else {
    croak("unhandled transaction type: $h->{trntype}");
  }

  my $a = "<STMTTRN>\n";
  $a .= "  <TRNTYPE>$h->{trntype}\n";
  $a .= "  <DTPOSTED>".datetime($h->{dtposted})."\n";
  $a .= "  <DTUSER>".datetime($h->{dtuser})."\n" if $h->{dtuser} > 0;
  $a .= "  <DTAVAIL>".datetime($h->{dtavail})."\n" if $h->{dtavail} > 0;
  $a .= "  <TRNAMT>$h->{trnamt}\n";
  $a .= "  <FITID>$h->{fitid}\n";
  $a .= "  <CORRECTFITID>$h->{correctfitid}\n" if $h->{correctfitid};
  $a .= "  <CORRECTACTION>$h->{correctaction}\n" if $h->{correctaction};
  $a .= "  <SRVRTID>$h->{srvrtid}\n" if $h->{srvrtid};
  $a .= "  <CHECKNUM>$h->{checknum}\n" if $h->{checknum};
  $a .= "  <REFNUM>$h->{refnum}\n" if $h->{refnum};
  $a .= "  <SIC>$h->{sic}\n" if $h->{sic};
  $a .= "  <PAYEEID>$h->{payeeid}\n" if $h->{payeeid};
  $a .= "  <NAME>$h->{name}\n" if $h->{name};
  $a .= "  <MEMO>$h->{memo}\n" if $h->{memo};
  $a .= "</STMTTRN>\n";
  return $a;
}

# create a string with a <BANKTRANLIST> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# dtstart - start for transaction in seconds since epoch.
# dtend - end of transaction in seconds since epoch.
# stmttrn - array reference containing hash references for STMTTRN().
#
# @return string.
sub BANKTRANLIST
{
  my ($h) = @_;
  croak("BANKTRANLIST parameter is not a hash reference") unless ref($h) eq 'HASH';
  croak("BANKTRANLIST::dtstart is invalid") unless $h->{dtstart} > 0;
  croak("BANKTRANLIST::dtend is invalid") unless $h->{dtend} > 0;
  croak("BANKTRANLIST::stmttrn is not an array reference") unless ref($h->{stmttrn}) eq 'ARRAY';

  my $a = "<BANKTRANLIST>\n";
  $a .= "  <DTSTART>" . datetime($h->{dtstart}) . "\n";
  $a .= "  <DTEND>" . datetime($h->{dtend}) . "\n";
  foreach my $r (@{$h->{stmttrn}}) {
    $a .= STMTTRN($r);
  }
  $a .= "</BANKTRANLIST>\n";
  return $a;
}

# create a string with a <LEDGERBAL> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# balamt - ledger balance amount.
# dtasof - balance date in seconds since epoch.
#
# @return string.
sub LEDGERBAL
{
  my ($h) = @_;
  croak("LEDGERBAL parameter is not a hash reference") unless ref($h) eq 'HASH';
  croak("LEDGERBAL::balamt not found") unless length($h->{balamt});
  # \todo check balamt for numeric value.
  croak("LEDGERBAL::dtasof not found") unless $h->{dtasof} > 0;
  my $a = "<LEDGERBAL>\n";
  $a .= "  <BALAMT>$h->{balamt}\n";
  $a .= "  <DTASOF>".datetime($h->{dtasof})."\n";
  $a .= "</LEDGERBAL>\n";
  return $a;
}

# create a string with a <AVAILBAL> tag.
# If the parameter is not a hash reference, the empty string is returned.
#
# the hash reference parameter needs to contain the following key/value pairs:
# balamt - ledger balance amount.
# dtasof - balance date in seconds since epoch.
#
# @return string.
sub AVAILBAL
{
  my ($h) = @_;
  return '' unless ref($h) eq 'HASH';
  croak("AVAILBAL::balamt not found") unless length($h->{balamt});
  # \todo check balamt for numeric value.
  croak("AVAILBAL::dtasof not found") unless $h->{dtasof} > 0;
  my $a = "<AVAILBAL>\n";
  $a .= "  <BALAMT>$h->{balamt}\n";
  $a .= "  <DTASOF>".datetime($h->{dtasof})."\n";
  $a .= "</AVAILBAL>\n";
  return $a;
}

# create a string with a <STMTRS> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# curdef - currency string.
# bankacctfrom - hash reference for BANKACCTFROM().
# banktranlist - hash reference for BANKTRANLIST().
# ledgerbal - hash reference for LEDGERBAL().
#
# the hash reference parameter can contain the following optional key/value pairs:
# availbal - hash reference for AVAILBAL().
#
# @return string.
sub STMTRS
{
  my ($h) = @_;
  croak("STMTRS parameter is not a hash reference") unless ref($h) eq 'HASH';
  croak("STMTRS::curdef not found") unless $h->{curdef};
  croak("STMTRS::bankacctfrom not a hash ref") unless ref($h->{bankacctfrom}) eq 'HASH';
  croak("STMTRS::banktranlist not a hash ref") unless ref($h->{banktranlist}) eq 'HASH';
  croak("STMTRS::ledgerbal not a hash ref") unless ref($h->{ledgerbal}) eq 'HASH';

  my $bankacctfrom = BANKACCTFROM($h->{bankacctfrom});
  my $banktranlist = BANKTRANLIST($h->{banktranlist});
  my $ledgerbal = LEDGERBAL($h->{ledgerbal});
  my $availbal = AVAILBAL($h->{availbal});
  my $a = <<"EOT";
<STMTRS>
  <CURDEF>$h->{curdef}
$bankacctfrom
$banktranlist
$ledgerbal
$availbal
</STMTRS>
EOT
  return $a;
}

# create a string with a <STMTTRNRS> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# stmtrs - hash reference for STMTRS().
#
# the hash reference parameter can contain the following optional key/value pairs:
# status - hash reference for STATUS().
# trnuid - transaction UUID string.
#
# @return string.
sub STMTTRNRS
{
  my ($h) = @_;

  my $status = STATUS($h->{status});

  my $trniud = $h->{trnuid};
  $trnuid = 0 unless $trnuid;

  my $stmtrs = STMTRS($h->{stmtrs});

  my $a = <<"EOT";
<STMTTRNRS>
  <TRNUID>$trnuid
$status
$stmtrs
</STMTTRNRS>
EOT
  return $a;
}

# create a string with a <BANKMSGSRS> tag.
#
# the hash reference parameter needs to contain the following key/value pairs:
# stmttrnrs - hash reference for STMTTRNRS()
#
# @return string.
sub BANKMSGSRSV1
{
  my ($h) = @_;
  croak("did not find BANKMSGSRSV1::stmttrnrs") unless ref($h->{stmttrnrs}) eq 'HASH';
  my $a = "<BANKMSGSRSV1>\n";
  $a .= STMTTRNRS($h->{stmttrnrs});
  $a .= "</BANKMSGSRSV1>\n";
  return $a;
}

# create a string with a <OFX> tag.
#
# the hash reference parameter can contain the following optional key/value pairs:
# signonmsgsrs - hash reference for SIGNONMSGSRSV1()
# bankmsgsrs - hash reference for BANKMSGSRSV1()
#
# @return string.
sub OFX
{
  my ($h) = @_;
  my $a = "<OFX>\n";
  $a .= SIGNONMSGSRSV1($h->{signonmsgsrs});
  $a .= BANKMSGSRSV1($h->{bankmsgsrs});
  $a .= "</OFX>\n";
  return $a;
}

1;
