#!/usr/bin/env perl
# @(#) A program to convert PayPal CSV files to IIF files to import into Quickbooks.
# $Id: csv2iif.pl 7554 2013-12-22 23:37:24Z doj $
#
# Copyright (c) 2013 Dirk Jagdmann <doj@cubic.org>
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
#    1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgement in the product documentation would be
#    appreciated but is not required.
#
#    2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
#
#    3. This notice may not be removed or altered from any source
#    distribution.
#
# For more details visit the website http://llg.cubic.org/csv2iif

use Text::CSV_XS;
use POSIX;
use strict;
use lib './';
use PayPal;
use PayPal::DB;
use Quickbooks::IIF;
use Time::Local;
use JSON;

# check command line
PayPal::DB::parse_command_line(@ARGV);
my $fn;
my $force = 0;
foreach my $arg (@ARGV) {
	help() if $arg eq '--help' || $arg eq '-h';
	$force = 1 if $arg eq '--force';
	if (-f $arg) {
		$fn = $arg;
		last;
	}
}
if (! defined($fn) && -f 'Download.csv')
{
	print STDERR "using Download.csv, since no filename was specified.\n";
	$fn = 'Download.csv';
}
help() unless -f $fn;

# copy CSV to PayPal database
my @txn; # transactions to convert to IIF
open(my $fh, "<", $fn) or die "could not open $fn: $!";
my $csv = Text::CSV_XS->new ({ binary => 1, allow_whitespace => 1 });
$csv->column_names(@{$csv->getline($fh)});
my $line = 1;
while (my $h = $csv->getline_hr($fh))
{
    ++$line;

    # check if this transaction has been converted before
    my $txn = $h->{'Transaction ID'};
    if (!$force && PayPal::DB::get_state($txn)) {
	print STDERR "skipping $txn from $h->{Date} converted ".asctime(localtime(PayPal::DB::get_state($txn)));
	next;
    }

    # rename keys
    $h->{GrossAmount} = $h->{Gross}; delete $h->{Gross};
    $h->{NetAmount} = $h->{Net}; delete $h->{Net};
    $h->{FeeAmount} = $h->{Fee}; delete $h->{Fee};
    $h->{Payer} = $h->{'From Email Address'}; delete $h->{'From Email Address'};
    $h->{Receiver} = $h->{'To Email Address'}; delete $h->{'To Email Address'};
    $h->{Title} = $h->{'Item Title'}; delete $h->{'Item Title'};
    $h->{PayerDisplayName} = $h->{Name}; delete $h->{Name};
    $h->{ParentTransactionID} = $h->{'Reference Txn ID'}; delete $h->{'Reference Txn ID'};

    # strip whitespace from keys
    foreach my $k (keys %$h) {
	my $kk = $k;
	$kk =~ s/\s+//g;	# strip whitespace from $kk
	if ($kk ne $k && !exists($h->{$kk})) {
	    $h->{$kk} = $h->{$k};
	    delete $h->{$k};
	}
    }

    # create timestamp
    my $day;
    my $mon;
    my $year;
    if ($h->{Date} =~ m!^(\d{1,2})/(\d{1,2})/(\d{4})$!) {
	$mon = $1;
	$day = $2;
	$year = $3;
    } else {
	die "$fn:$line: unknown Date format: $h->{Date}";
    }

    my $hour;
    my $min;
    my $sec;
    if ($h->{Time} =~ /^(\d\d):(\d\d):(\d\d)$/) {
	$hour = $1;
	$min = $2;
	$sec = $3;
    } else {
	die "$fn:$line: unknown Time format: $h->{Time}";
    }

    my $ts = timegm($sec, $min, $hour, $day, $mon - 1, $year);

    our %timezone2offset = (
			    GMT => 0, # Greenwich Mean Time
			    UTC => 0, # Universal Time Coordinated
			    CET =>  +1 * 3600, # Central European Time
			    CEST => +2 * 3600, # Central European Summer Time
			    HST => -10 * 3600, # Hawaii Standard Time
			    HDT =>  -9 * 3600, # Hawaii Daylight savings Time
			    AKST => -9 * 3600, # Alaska Standard Time
			    AKDT => -8 * 3600, # Alaska Daylight savings Time
			    PST => -8 * 3600, # Pacific Standard Time
			    PDT => -7 * 3600, # Pacific Daylight savings Time
			    MST => -7 * 3600, # Mountain Standard Time
			    MDT => -6 * 3600, # Mountain Daylight savings Time
			    CST => -6 * 3600, # Central Standard Time
			    CDT => -5 * 3600, # Central Daylight savings Time
			    EST => -5 * 3600, # Eastern Standard Time
			    EDT => -4 * 3600, # Eastern Daylight savings Time
			   );
    die "unknown Time Zone: $h->{TimeZone}" unless exists $timezone2offset{$h->{TimeZone}};
    $ts += - $timezone2offset{$h->{TimeZone}};

    $h->{Timestamp} = PayPal::ts2str($ts);

    # done preprocessing CSV
    PayPal::DB::save_txn($h) or die "$fn:$line: could not save PayPal transaction";
    push(@txn, $txn);
}
$csv->eof() or $csv->error_diag;
close($fh) or die "could not close $fn: $!";

my $iif = $fn;
$iif =~ s/\.csv$//i;
$iif .= '.iif';
my $res = PayPal::iif($iif, @txn);
die "could not save $iif" if ! defined($res);
if ($res == 0) {
    print "No transactions saved.\n";
    exit 0;
}
print "saved $res transactions to $iif\n";

# add new transactions to the state hash
PayPal::DB::set_states(time(), @txn) or die "could not set txn state";
exit 0;

sub help
{
    my $prog = 'csv2iif.pl';
  # when changing the help text, also modify the Command Line Details section in README.txt
  print<<"EOT";
usage: $prog
  Convert transactions from Download.csv into Download.iif
  Transactions which have been converted previously are skipped.

usage: $prog --force
  Convert transactions from Download.csv into Download.iif
  All transactions will be converted including those from a previous
  conversion.

usage: $prog file.csv
  Convert transactions from file.csv into file.iif
  Transactions which have been converted previously are skipped.

usage: $prog --force file.csv
  Convert transactions from file.csv into file.iif
  All transactions will be converted including those from a previous
  conversion.
EOT
    PayPal::DB::command_line_help($prog);
    exit 1;
}
