#!/usr/bin/env perl
# @(#) A program to download PayPal transactions and save them as OFX files.
#
# Copyright (c) 2015 Dirk Jagdmann <doj@cubic.org>
#
# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
#    1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgement in the product documentation would be
#    appreciated but is not required.
#
#    2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
#
#    3. This notice may not be removed or altered from any source
#    distribution.
#
# For more details visit the website http://llg.cubic.org/csv2iif

use strict;
use OFX::OFX1;
use Business::PayPal::API qw( GetTransactionDetails TransactionSearch GetBalance );
use PayPal;
use PayPal::DB;
use Time::Local;
use JSON;
use POSIX;

##############################################################################
# command line
##############################################################################

my $ofx = 'PayPal.ofx';
my $do_help = 0;

my $end_ts = time();
my $start_ts = $end_ts - 86400*10;
if (PayPal::DB::get_state('last_paypal_API_sync') > 0)
{
    $start_ts = PayPal::DB::get_state('last_paypal_API_sync') - 1;
    die "strange start ts $start_ts" unless $start_ts > 100;
}

# parse command line
PayPal::DB::parse_command_line(@ARGV);
foreach my $arg (@ARGV) {
    if ($ofx eq 'fetch me') {
	$ofx = $arg;
	help() unless length($ofx);
	next
    }
    if ($arg eq '-o') {
	$ofx = 'fetch me';
	next
    }

    if ($start_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --start, needs to be 'YYYY-MM-DD'";
	}
	$start_ts = timelocal(0, 0, 0, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --start" unless $start_ts > 100;
    }
    if ($arg eq '--start') {
	$start_ts = 'fetch me';
	next;
    }

    if ($end_ts eq 'fetch me') {
	unless($arg =~ /^(\d{4})-(\d\d)-(\d\d)$/) {
	    die "invalid day specification ($arg) for --end, needs to be 'YYYY-MM-DD'";
	}
	$end_ts = timelocal(59, 59, 23, $3, $2 - 1, $1);
	die "invalid day specification ($arg) for --end" unless $end_ts > 100;
    }
    if ($arg eq '--end') {
	$end_ts = 'fetch me';
	next;
    }

    if ($arg eq '--lastsync') {
	my $ts = PayPal::DB::get_state('last_paypal_API_sync');
	if ($ts > 0) {
	    print POSIX::ctime($ts);
	} else {
	    print "not synced with PayPal yet.\n";
	}
	exit 0;
    }

    if ($arg eq '--help' || $arg eq '-h') {
	$do_help = 1;
	next;
    }
}

help() if $start_ts eq 'fetch me';
help() if $end_ts eq 'fetch me';
die "start day need to be before end day" if $start_ts > $end_ts;

help() if $do_help;

##############################################################################
# transaction struct
##############################################################################

my $fi = {
	  org => 'PAYPAL',
	  fi => '1000'
};
my $signonmsgsrs = {
		     fi => $fi,
};
my $bankacctfrom = {
		    bankid => '031101169',
		    acctid => '123456789',
		    accttype => 'CHECKING',
};
my $stmttrn = [ ];
my $banktranlist = {
		    dtstart => $start_ts,
		    dtend => $end_ts,
		    stmttrn => $stmttrn,
};
my $ledgerbal = {
		 balamt => 999999999, # will be set from GetBalance()
		 dtasof => $end_ts,   # will be set from GetBalance()
};
my $stmtrs = {
	      curdef => $configuration::paypal_currency,
	      bankacctfrom => $bankacctfrom,
	      banktranlist => $banktranlist,
	      ledgerbal => $ledgerbal,
	      availbal => $ledgerbal,
};
my $stmttrnrs = {
		 stmtrs => $stmtrs,
};
my $bankmsgsrs = {
		  stmttrnrs => $stmttrnrs
};

##############################################################################
# query PayPal
##############################################################################

my $pp = new Business::PayPal::API(
				   Username   => $configuration::paypal_username,
				   Password   => $configuration::paypal_password,
				   Signature  => $configuration::paypal_signature,
				   sandbox    => $configuration::paypal_sandbox
				  ) or die "could not init PayPal API";

my $start_str = PayPal::ts2str($start_ts);
my $end_str = PayPal::ts2str($end_ts);
print "getting PayPal transactions from $start_str to $end_str ...\n";

$| = 1; # auto flush STDOUT
my @txn;
my $total = 0;
while(1)
{
    $end_str = PayPal::ts2str($end_ts);
    #print "# requesting $start_str to $end_str\n\n";
    my $a = $pp->TransactionSearch( StartDate => $start_str, EndDate => $end_str );

    my $z = 0;
    my $last_ts = PayPal::str2ts('2038-01-01T00:00:00Z'); # \todo fix sometime in 2037
    foreach my $h (@$a) {
	++$z;
	++$total;
	my $txn = $h->{TransactionID};
#	if (PayPal::DB::get_state($txn)) {
#	    print STDERR "skipping $txn from $h->{Timestamp} converted ".asctime(localtime(PayPal::DB::get_state($txn)));
#	    next;
#	}

	my $ts = PayPal::str2ts($h->{Timestamp});
	die "time stamp is not less than previous time stamp" unless $ts <= $last_ts;
	$last_ts = $ts;

	my %t;
	my $t = PayPal::DB::get_txn($txn);
	if (ref($t) eq 'HASH') {
	    %t = %$t;
	} else {
	    %t = $pp->GetTransactionDetails( TransactionID => $txn );
	}
	# merge $h into %t
	@t{keys %$h} = values %$h;

	#print STDERR "##########################################################\n";
	#print STDERR "# $txn\n";
	#dump_txn(\%t);

	push(@txn, $txn) if $txn;
	PayPal::DB::save_txn(\%t) or die "$txn could not save PayPal transaction";

	print "\r$total ";
    }

    last if $z < 100;
    die "more than 100 transaction on ".PayPal::ts2str($last_ts) if $end_ts == $last_ts;
    $end_ts = $last_ts + 1; # since there may be multiple transactions on any time stamp, move end_ts out 1 second to catch them all
}
print "\rfetched $total transactions from PayPal.\n";

my %t = $pp->GetBalance();
if ($t{Ack} eq 'Success') {
  $ledgerbal->{balamt} = $t{Balance};
  $ledgerbal->{dtasof} = PayPal::str2ts($t{BalanceTimeStamp});
} else {
  print "Could not make GetBalance() call:\n";
  while(my ($k,$v) = each(%t)) {
    print "$k\t$v\n";
  }
  exit 1;
}

sub PII_Name
{
  my ($h) = @_;
  return '' unless $h;
  return $h if ref($h) eq 'SCALAR';
  return join(', ', @$h) if ref($h) eq 'ARRAY';
  return "ref=".ref($h);
}

foreach my $txn (@txn) {
  my $t = PayPal::DB::get_txn($txn);
  die "could not load transaction $txn from disk" unless ref($t) eq 'HASH';
  my $h = { name => "$t->{PayerDisplayName} $t->{Payer} ".PII_Name($t->{PII_Name}),
	    dtposted => PayPal::str2ts($t->{Timestamp}),
	    fitid => $t->{TransactionID},
	    trnamt => $t->{NetAmount},
	    memo => "$t->{Type}: $t->{Memo}",
	  };

  if ($t->{Type} eq "Payment") {
    if ($t->{NetAmount} < 0.0) {
      dump_txn($t);
      die "Type Payment expects positive NetAmount";
    }
    $h->{trntype} = 'DEP';
  } elsif ($t->{Type} eq "Transfer") {
    if ($t->{NetAmount} > 0.0) {
      dump_txn($t);
      die "Type Transfer expects negative NetAmount";
    }
    $h->{trntype} = 'XFER';
  } elsif ($t->{Type} eq "Fee Reversal") {
    if ($t->{NetAmount} < 0.0) {
      dump_txn($t);
      die "Type Fee Reversal expects positive NetAmount";
    }
    $h->{trntype} = 'CREDIT';
    $h->{name} = $t->{Type};
  } elsif ($t->{Type} eq "Refund") {
    if ($t->{NetAmount} > 0.0) {
      dump_txn($t);
      die "Type Refund expects negative NetAmount";
    }
    $h->{trntype} = 'DEBIT';
  } else {
      dump_txn($t);
      die "unhandled Type: '$t->{Type}'";
  }

  push(@$stmttrn, $h);
}

PayPal::DB::set_state('last_paypal_API_sync', $end_ts) or die "could not set sync time state";
#PayPal::DB::set_states($end_ts, @txn) or die "could not set txn state";

##############################################################################
# output
##############################################################################

open(A, '>', $ofx) or die "could not open for write: $ofx";
print A OFX::OFX1::header();
print A OFX::OFX1::OFX({ signonmsgsrs => $signonmsgsrs, bankmsgsrs => $bankmsgsrs });
close(A);

print "saved ".scalar(@$stmttrn)." transactions to $ofx\n";
exit 0;

##############################################################################

sub help
{
    my $prog = 'paypal2ofx1.pl';
    print<<"EOT";
usage: $prog
  Fetch transactions from PayPal and convert to $ofx

usage: $prog -o out.ofx
  Fetch transactions from PayPal and convert to out.ofx

usage: $prog --start 'YYYY-MM-DD'
  Fetch transactions from PayPal starting at midnight from the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --end 'YYYY-MM-DD'
  Fetch transactions from PayPal until 23:59:59 at the day specified.
  Midnight is interpreted in your local time zone. Remember that PayPal uses
  UTC, so the timestamps displayed may have an offset.

usage: $prog --lastsync
  Show timestamp of last sync with PayPal.
EOT
    PayPal::DB::command_line_help($prog);
    exit 1;
}

sub dump_txn
{
  my ($txn) = @_;
  die "dump_txn parameter is not a hash reference" unless ref($txn) eq 'HASH';
  print STDERR to_json($txn, {utf8 => 1, pretty => 1}) . "\n";
}
